package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.testing.base.ApiTestConfig;
import be.iminds.ilabt.jfed.testing.base.ApiTestConfigProperty;
import be.iminds.ilabt.jfed.testing.shared.NodeLoginTestStepConfig;
import be.iminds.ilabt.jfed.testing.shared.Proxy;
import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.security.PrivateKey;
import java.security.PublicKey;

@JsonInclude(JsonInclude.Include.NON_DEFAULT)
@JsonIgnoreProperties(value={"@context"})
public class ESpecTestConfig implements ApiTestConfig {
    private static final Logger LOG = LoggerFactory.getLogger(ESpecTestConfig.class);

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    @Nonnull private final be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy;
    @Nonnull private final String amVersion;
    @Nonnull private final SSHKeys sshKeys;
    @Nonnull private final Slice slice;
    @Nonnull private final ESpec eSpec;
    @Nonnull private final Provision provision;
    @Nonnull private final WaitForReady waitForReady;
    @Nonnull private final Cleanup cleanup;

    public ESpecTestConfig() {
        this.amConnectionProxy = new be.iminds.ilabt.jfed.testing.shared.Proxy();
        this.amVersion = "auto";
        this.sshKeys = new SSHKeys();

        this.slice = new Slice();
        this.eSpec = new ESpec();
        this.provision = new Provision();
        this.waitForReady = new WaitForReady();
        this.cleanup = new Cleanup();
    }

    @JsonCreator
    public ESpecTestConfig(@Nullable @JsonProperty("amConnectionProxy") be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy,
                           @Nullable @JsonProperty("amVersion") String amVersion,
                           @Nullable @JsonProperty("sshKeys") SSHKeys sshKeys,
                           @Nullable @JsonProperty("slice") Slice slice,
                           @Nullable @JsonProperty("eSpec") ESpec eSpec,
                           @Nullable @JsonProperty("provision") Provision provision,
                           @Nullable @JsonProperty("waitForReady") WaitForReady waitForReady,
                           @Nullable @JsonProperty("cleanup") Cleanup cleanup) {
        this.amConnectionProxy = amConnectionProxy == null ? new ESpecTestConfig().getAmConnectionProxy() : amConnectionProxy;
        this.amVersion = amVersion == null ? new ESpecTestConfig().getAmVersion() : amVersion;
        this.sshKeys = sshKeys == null ? new ESpecTestConfig().getSshKeys() : sshKeys;
        this.slice = slice == null ? new Slice() : slice;
        this.eSpec = eSpec == null ? new ESpec() : eSpec;
        this.provision = provision == null ? new Provision() : provision;
        this.waitForReady = waitForReady == null ? new WaitForReady() : waitForReady;
        this.cleanup = cleanup == null ? new Cleanup() : cleanup;
    }

    @Override
    @JsonIgnore
    public boolean isValid() {
        boolean res = getAmConnectionProxy().isValid();
        res = res && getSshKeys().isValid();
        boolean amVersionValid = getAmVersion().equals("2") || getAmVersion().equals("3") || getAmVersion().equals("auto");
        if (!amVersionValid)
            throw new IllegalArgumentException("Config Error: Unsupported amVersion: \""+getAmVersion()+"\"");
//        res = res && amVersionValid;

        res = res && getSlice().isValid();
        res = res && getESpec().isValid();
        res = res && getProvision().isValid();
        res = res && getWaitForReady().isValid();
        res = res && getCleanup().isValid();
        if (!getProvision().isEnabled() && !getWaitForReady().isEnabled())
            throw new IllegalArgumentException("Config Error: if the Provision step is disabled, waitForReady should be disabled too.");

        return res;
    }

    @JsonProperty
    @Nonnull
    public be.iminds.ilabt.jfed.testing.shared.Proxy getAmConnectionProxy() {
        return amConnectionProxy;
    }

    @JsonProperty
    @Nonnull
    public String getAmVersion() {
        return amVersion;
    }

    @JsonProperty
    @Nonnull
    public SSHKeys getSshKeys() {
        return sshKeys;
    }


    @JsonProperty
    @Nonnull
    public Slice getSlice() {
        return slice;
    }

    @JsonProperty("eSpec")
    @Nonnull
    public ESpec getESpec() {
        return eSpec;
    }

    @JsonProperty
    @Nonnull
    public Provision getProvision() {
        return provision;
    }

    @JsonProperty
    @Nonnull
    public WaitForReady getWaitForReady() {
        return waitForReady;
    }

    @JsonProperty
    @Nonnull
    public Cleanup getCleanup() {
        return cleanup;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof ESpecTestConfig)) return false;

        ESpecTestConfig that = (ESpecTestConfig) o;

        if (!amConnectionProxy.equals(that.amConnectionProxy)) return false;
        if (!amVersion.equals(that.amVersion)) return false;
        if (!sshKeys.equals(that.sshKeys)) return false;
        if (!slice.equals(that.slice)) return false;
        if (!eSpec.equals(that.eSpec)) return false;
        if (!provision.equals(that.provision)) return false;
        if (!waitForReady.equals(that.waitForReady)) return false;
        return cleanup.equals(that.cleanup);
    }

    @Override
    public int hashCode() {
        int result = amConnectionProxy.hashCode();
        result = 31 * result + amVersion.hashCode();
        result = 31 * result + sshKeys.hashCode();
        result = 31 * result + slice.hashCode();
        result = 31 * result + eSpec.hashCode();
        result = 31 * result + provision.hashCode();
        result = 31 * result + waitForReady.hashCode();
        result = 31 * result + cleanup.hashCode();
        return result;
    }

    @Nonnull
    @Override
    @JsonIgnore
    public String stringValue() {
        try {
            return MAPPER.writeValueAsString(this);
        } catch (JsonProcessingException e) {
            throw new RuntimeException("Failed to covert "+getClass().getName()+" to String", e);
        }
    }

    @Override
    public String toString() {
        return stringValue();
    }

    public enum SSHKeySource { TEST_USER, RANDOM };

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class SSHKeys {
        @Nonnull private final SSHKeySource source;

        //maybe future, but not sure how to do this securily from fedmon:
        // , fixedContent, fixedFile
//        @Nonnull private final String publicKeyContent;
//        @Nonnull private final String privateKeyContent;
//        @Nonnull private final String publicKeyFile;
//        @Nonnull private final String privateKeyFile;
//        @Nonnull private final String privateKeyPassword;


        public SSHKeys() {
            this.source = SSHKeySource.RANDOM;
        }

        @JsonCreator
        public SSHKeys(@Nullable @JsonProperty("source") SSHKeySource source) {
            this.source = source == null ? new SSHKeys().getSource() : source;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }

        @JsonProperty
        @Nonnull
        public SSHKeySource getSource() {
            return source;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof SSHKeys)) return false;

            SSHKeys sshKeys = (SSHKeys) o;

            return source == sshKeys.source;
        }

        @Override
        public int hashCode() {
            return source.hashCode();
        }
    }

    public enum ESpecSource {PROVIDE_ARCHIVE_URL, PROVIDE_ARCHIVE_FILE, PROVIDE_DIR, PROVIDE_GIT_REPO_DIR };
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class ESpec {
        @Nonnull private final ESpecSource source;
        @Nullable private final String providedContentSource; //url, file or content as specified in source

        public ESpec() {
            this.source = ESpecSource.PROVIDE_DIR;
            this.providedContentSource = null;
        }

        @JsonCreator
        public ESpec(@Nullable @JsonProperty("source") ESpecSource source,
                     @Nullable @JsonProperty("providedContentSource") String providedContentSource) {
            this.source = source == null ? new ESpec().getSource() : source;
            this.providedContentSource = providedContentSource;
        }

        @JsonIgnore
        public boolean isValid() {
            switch (source) {
                case PROVIDE_ARCHIVE_FILE:
                case PROVIDE_ARCHIVE_URL:
                case PROVIDE_DIR: {
                    if (providedContentSource == null)
                        throw new IllegalArgumentException("Config Error: rspec.providedContentSource may not be null if espec.source == "+source);
                    return true;
                }
                default:
                    throw new IllegalArgumentException("Config Error: unsupported espec.source == "+source);
            }
        }

        @JsonProperty
        @Nonnull
        public ESpecSource getSource() {
            return source;
        }

        @JsonProperty
        @Nullable
        public String getProvidedContentSource() {
            return providedContentSource;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof ESpec)) return false;

            ESpec eSpec = (ESpec) o;

            if (source != eSpec.source) return false;
            return providedContentSource != null ? providedContentSource.equals(eSpec.providedContentSource) : eSpec.providedContentSource == null;
        }

        @Override
        public int hashCode() {
            int result = source.hashCode();
            result = 31 * result + (providedContentSource != null ? providedContentSource.hashCode() : 0);
            return result;
        }
    }

    public enum ProjectSource { NONE, AUTO, PROVIDED }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Slice {
        @Nullable private final String sliceName;
        private final boolean failOnExistingSlice;
        private final boolean failOnNonExistingSlice;
        private final long expireTimeMin;
        private final boolean renewExistingSliceIfNeeded;

        @Nonnull private final ProjectSource projectSource;
        private final boolean failIfNoProject; //if projectSource is AUTO or PROVIDED
        @Nullable private final String project; //if projectSource is PROVIDED

        public Slice() {
            this.sliceName = null;
            this.failOnExistingSlice = true;
            this.failOnNonExistingSlice = false;
            this.expireTimeMin = 60;
            this.renewExistingSliceIfNeeded = true;
            this.projectSource = ProjectSource.AUTO;
            this.failIfNoProject = false;
            this.project = null;
        }

        @JsonCreator
        public Slice(@Nullable @JsonProperty("sliceName") String sliceName,
                     @Nullable @JsonProperty("failOnExistingSlice") Boolean failOnExistingSlice,
                     @Nullable @JsonProperty("failOnNonExistingSlice") Boolean failOnNonExistingSlice,
                     @Nullable @JsonProperty("expireTimeMin") Long expireTimeMin,
                     @Nullable @JsonProperty("renewExistingSliceIfNeeded") Boolean renewExistingSliceIfNeeded,
                     @Nullable @JsonProperty("projectSource") ProjectSource projectSource,
                     @Nullable @JsonProperty("failIfNoProject") Boolean failIfNoProject,
                     @Nullable @JsonProperty("project") String project) {
            this.sliceName = sliceName == null ? new Slice().getSliceName() : sliceName;
            this.failOnExistingSlice = failOnExistingSlice == null ? new Slice().getFailOnExistingSlice() : failOnExistingSlice;
            this.failOnNonExistingSlice = failOnNonExistingSlice == null ? new Slice().getFailOnNonExistingSlice() : failOnNonExistingSlice;
            this.expireTimeMin = expireTimeMin == null ? new Slice().getExpireTimeMin() : expireTimeMin;
            this.renewExistingSliceIfNeeded = renewExistingSliceIfNeeded == null ? new Slice().getRenewExistingSliceIfNeeded() : renewExistingSliceIfNeeded;
            this.projectSource = projectSource == null ? new Slice().getProjectSource() : projectSource;
            this.failIfNoProject = failIfNoProject == null ? new Slice().getFailIfNoProject() : failIfNoProject;
            this.project = project == null ? new Slice().getProject() : project;
        }

        @JsonIgnore
        public boolean isValid() {
            if (expireTimeMin <= 0)
                throw new IllegalArgumentException("Config Error: slice.expireTimeMin <= makes no sense: "+expireTimeMin);
            if (projectSource == ProjectSource.PROVIDED && project == null && failIfNoProject)
                throw new IllegalArgumentException("Config Error: Combination of failIfNoProject=true and project==null and projectSource == ProjectSource.PROVIDED will always fail");
            if (projectSource == ProjectSource.NONE && failIfNoProject)
                throw new IllegalArgumentException("Config Error: Combination of failIfNoProject=true and projectSource == ProjectSource.NONE will always fail");
            if (failOnExistingSlice && failOnNonExistingSlice)
                throw new IllegalArgumentException("Config Error: failOnExistingSlice and failOnNonExistingSlice should not both be true");

            return true;
        }

        @Nullable
        public String getSliceName() {
            return sliceName;
        }

        public boolean getFailOnExistingSlice() {
            return failOnExistingSlice;
        }

        public boolean getFailOnNonExistingSlice() {
            return failOnNonExistingSlice;
        }

        public long getExpireTimeMin() {
            return expireTimeMin;
        }

        public boolean getRenewExistingSliceIfNeeded() {
            return renewExistingSliceIfNeeded;
        }

        @Nonnull
        public ProjectSource getProjectSource() {
            return projectSource;
        }

        public boolean getFailIfNoProject() {
            return failIfNoProject;
        }

        @Nullable
        public String getProject() {
            return project;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Slice)) return false;

            Slice slice = (Slice) o;

            if (failOnExistingSlice != slice.failOnExistingSlice) return false;
            if (failOnNonExistingSlice != slice.failOnNonExistingSlice) return false;
            if (expireTimeMin != slice.expireTimeMin) return false;
            if (renewExistingSliceIfNeeded != slice.renewExistingSliceIfNeeded) return false;
            if (failIfNoProject != slice.failIfNoProject) return false;
            if (sliceName != null ? !sliceName.equals(slice.sliceName) : slice.sliceName != null) return false;
            if (projectSource != slice.projectSource) return false;
            return project != null ? project.equals(slice.project) : slice.project == null;
        }

        @Override
        public int hashCode() {
            int result = sliceName != null ? sliceName.hashCode() : 0;
            result = 31 * result + (failOnExistingSlice ? 1 : 0);
            result = 31 * result + (failOnNonExistingSlice ? 1 : 0);
            result = 31 * result + (int) (expireTimeMin ^ (expireTimeMin >>> 32));
            result = 31 * result + (renewExistingSliceIfNeeded ? 1 : 0);
            result = 31 * result + projectSource.hashCode();
            result = 31 * result + (failIfNoProject ? 1 : 0);
            result = 31 * result + (project != null ? project.hashCode() : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Provision {
        //Allocate, Provision, POA    (skipped if requestRspec method is "already_povisioned")
//        private final long provisionExpireTimeMin;
        private final boolean useGeniEndTimeOptionForCreateSliver;
        private final boolean enabled;

        public Provision() {
//            this.provisionExpireTimeMin = 25;
            this.useGeniEndTimeOptionForCreateSliver = false;
            this.enabled = true;
        }
        @JsonCreator
        public Provision(
//                @Nullable @JsonProperty("provisionExpireTimeMin") Long provisionExpireTimeMin,
                @Nullable @JsonProperty("useGeniEndTimeOptionForCreateSliver") Boolean useGeniEndTimeOptionForCreateSliver,
                @Nullable @JsonProperty("enabled") Boolean enabled
        ) {
//            this.provisionExpireTimeMin = provisionExpireTimeMin == null ? new Provision().getProvisionExpireTimeMin() : provisionExpireTimeMin;
            this.useGeniEndTimeOptionForCreateSliver = useGeniEndTimeOptionForCreateSliver == null ? new Provision().getUseGeniEndTimeOptionForCreateSliver() : useGeniEndTimeOptionForCreateSliver;
            this.enabled = enabled == null ? new Provision().isEnabled() : enabled;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }

//        @JsonProperty
//        public long getProvisionExpireTimeMin() {
//            return provisionExpireTimeMin;
//        }

        @JsonProperty
        public boolean getUseGeniEndTimeOptionForCreateSliver() {
            return useGeniEndTimeOptionForCreateSliver;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Provision)) return false;

            Provision provision = (Provision) o;

            if (useGeniEndTimeOptionForCreateSliver != provision.useGeniEndTimeOptionForCreateSliver) return false;
            return enabled == provision.enabled;
        }

        @Override
        public int hashCode() {
            int result = (useGeniEndTimeOptionForCreateSliver ? 1 : 0);
            result = 31 * result + (enabled ? 1 : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class WaitForReady {
        private final long maxTimeMin;
        private final boolean skipFinalManifestCheck;
        private final boolean requireNodesInManifest;
        private final long extraWaitTimeSeconds;
        private final boolean enabled;

        public WaitForReady() {
            this.maxTimeMin = 20;
            this.skipFinalManifestCheck = false;
            this.requireNodesInManifest = true;
            this.extraWaitTimeSeconds = 0;
            this.enabled = true;
        }
        @JsonCreator
        public WaitForReady(@Nullable @JsonProperty("maxTimeMin") Long maxTimeMin,
                            @Nullable @JsonProperty("skipFinalManifestCheck") Boolean skipFinalManifestCheck,
                            @Nullable @JsonProperty("requireNodesInManifest") Boolean requireNodesInManifest,
                            @Nullable @JsonProperty("extraWaitTimeSeconds") Long extraWaitTimeSeconds,
                            @Nullable @JsonProperty("enabled") Boolean enabled) {
            this.maxTimeMin = maxTimeMin == null ? new WaitForReady().getMaxTimeMin() : maxTimeMin;
            this.skipFinalManifestCheck = skipFinalManifestCheck == null ? new WaitForReady().getSkipFinalManifestCheck() : skipFinalManifestCheck;
            this.requireNodesInManifest = requireNodesInManifest == null ? new WaitForReady().getRequireNodesInManifest() : requireNodesInManifest;
            this.extraWaitTimeSeconds = extraWaitTimeSeconds == null ? new WaitForReady().getExtraWaitTimeSeconds() : extraWaitTimeSeconds;
            this.enabled = enabled == null ? new WaitForReady().isEnabled() : enabled;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!enabled)
                return true;

            if (maxTimeMin < 0)
                throw new IllegalArgumentException("Config Error: waitForReady.maxTimeMin < 0 makes no sense");

            return true;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Skip the detailed manifest test in the \"checkManifest\" step? If skipped, very basic testing of the manifest is still done.")
        public boolean getSkipFinalManifestCheck() {
            return skipFinalManifestCheck;
        }


        @JsonProperty
        public long getMaxTimeMin() {
            return maxTimeMin;
        }

        @JsonProperty
        public boolean getRequireNodesInManifest() {
            return requireNodesInManifest;
        }

        @JsonProperty
        public long getExtraWaitTimeSeconds() {
            return extraWaitTimeSeconds;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof WaitForReady)) return false;

            WaitForReady that = (WaitForReady) o;

            if (maxTimeMin != that.maxTimeMin) return false;
            if (skipFinalManifestCheck != that.skipFinalManifestCheck) return false;
            if (requireNodesInManifest != that.requireNodesInManifest) return false;
            if (extraWaitTimeSeconds != that.extraWaitTimeSeconds) return false;
            return enabled == that.enabled;
        }

        @Override
        public int hashCode() {
            int result = (int) (maxTimeMin ^ (maxTimeMin >>> 32));
            result = 31 * result + (skipFinalManifestCheck ? 1 : 0);
            result = 31 * result + (requireNodesInManifest ? 1 : 0);
            result = 31 * result + (int) (extraWaitTimeSeconds ^ (extraWaitTimeSeconds >>> 32));
            result = 31 * result + (enabled ? 1 : 0);
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Cleanup {
        private final boolean enabled;
        private final boolean ignoreErrors;
        private final long maxWaitTimeMinutes;

        public Cleanup() {
            this.enabled = true;
            this.ignoreErrors = false;
            this.maxWaitTimeMinutes = 5;
        }

        @JsonCreator
        public Cleanup(@Nullable @JsonProperty("enabled") Boolean enabled,
                       @Nullable @JsonProperty("ignoreErrors") Boolean ignoreErrors,
                       @Nullable @JsonProperty("maxWaitTimeMinutes") Long maxWaitTimeMinutes) {
            this.enabled = enabled == null ? new Cleanup().isEnabled() : enabled;
            this.ignoreErrors = ignoreErrors == null ? new Cleanup().getIgnoreErrors() : ignoreErrors;
            this.maxWaitTimeMinutes = maxWaitTimeMinutes == null ? new Cleanup().getMaxWaitTimeMinutes() : maxWaitTimeMinutes;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!enabled)
                return true;

            if (maxWaitTimeMinutes < 0)
                throw new IllegalArgumentException("Config Error: cleanup.maxWaitTimeMinutes < 0 makes no sense");

            return true;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @JsonProperty
        public boolean getIgnoreErrors() {
            return ignoreErrors;
        }

        @JsonProperty
        public long getMaxWaitTimeMinutes() {
            return maxWaitTimeMinutes;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Cleanup)) return false;

            Cleanup cleanup = (Cleanup) o;

            if (enabled != cleanup.enabled) return false;
            if (ignoreErrors != cleanup.ignoreErrors) return false;
            return maxWaitTimeMinutes == cleanup.maxWaitTimeMinutes;
        }

        @Override
        public int hashCode() {
            int result = (enabled ? 1 : 0);
            result = 31 * result + (ignoreErrors ? 1 : 0);
            result = 31 * result + (int) (maxWaitTimeMinutes ^ (maxWaitTimeMinutes >>> 32));
            return result;
        }
    }


    public NodeLoginTestStepConfig generateNodeLoginTestStepConfig(@Nonnull GeniUserProvider geniUserProvider) {
        NodeLoginTestStepConfig.SSHKeySource sshKeySource;
        PublicKey publicKey;
        PrivateKey privateKey;
        switch (getSshKeys().getSource()) {
            case RANDOM: {
                sshKeySource = NodeLoginTestStepConfig.SSHKeySource.RANDOM;
                publicKey = null;
                privateKey = null;
                break;
            }
            case TEST_USER: {
                sshKeySource = NodeLoginTestStepConfig.SSHKeySource.PROVIDED;
                assert geniUserProvider.isUserLoggedIn();
                GeniUser user = geniUserProvider.getLoggedInGeniUser();
                assert user != null;
                publicKey = user.getPublicKey();
                privateKey = user.getPrivateKey();
                break;
            }
            default: throw new IllegalArgumentException("Unsupported ssh keys source: "+getSshKeys().getSource());
        }


        return new NodeLoginTestStepConfig(sshKeySource,
                publicKey,
                privateKey,
                false,
                new Proxy(), //proxy OFF
                2*60 * 1000l,
                null,
                null);
    }
}
